#include <iostream>

class A_impl {
protected:
	int _value;
public:
	A_impl();
	~A_impl();
	virtual void print();
	virtual int value();
	virtual void value(int newValue);
};

/** representation of the smartwrapper **/
class ARep {
private:
	int refCnt;
	A_impl *a;

protected:
	ARep() : a(0),nextFree(0),refCnt(1) { }
	static ARep *aRepStore;

public:
	static ARep *alloc();
	ARep *nextFree;

	inline void create() { if(!a) a = new A_impl(); }
	inline A_impl *get() { create(); return a; }
	inline void copy() { ++refCnt; }
	inline void release() {
		refCnt--;
		if(!refCnt) {
			if(a) delete a;

			/** put into freelist and initialize for next use **/
			a = 0;
			refCnt = 1;
			nextFree = aRepStore;
			aRepStore = this;
		}
	}
};

class A {
protected:
	ARep *rep;

public:
	A() : rep(ARep::alloc()) {
	}
	A(const A& copyFrom) : rep(copyFrom.rep) {
		rep->copy();
	}
	inline A& operator=(const A& assignFrom) {
		rep->release();
		rep = assignFrom.rep;
		rep->copy();
		return *this;
	}
	~A() {
		rep->release();
	}
	/* this would of course be the easier way for forwarding
	inline A_impl* operator->() { return rep->get(); }
	*/

	/** forwarded functions **/
	inline void print() {
		rep->get()->print();
	}
	inline void value(int v) {
		rep->get()->value(v);
	}
	inline int value() {
		return rep->get()->value();
	}
};

extern void set5(A x);
extern void set5(A_impl& x);
